@@ -1,16 +1,38 @@
-/*  syspred.pl,v 1.19 1995/08/16 11:49:24 jan Exp
+/*  $Id: syspred.pl,v 1.66 2004/09/30 10:41:35 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Prolog system predicate definitions
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2004, University of Amsterdam
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License
+    as published by the Free Software Foundation; either version 2
+    of the License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+    As a special exception, if you link this library with other files,
+    compiled with a Free Software compiler, to produce an executable, this
+    library does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however
+    invalidate any other reasons why the executable file might be covered by
+    the GNU General Public License.
 */
 
 :- module($syspreds,
 	[ leash/1
 	, visible/1
 	, style_check/1
-	, please/3
 	, (spy)/1
 	, (nospy)/1
 	, trace/1
@@ -19,20 +41,15 @@
 	, debugging/0
 	, concat_atom/2
 	, term_to_atom/2
-	, atom_to_term/3
 	, int_to_atom/2
-	, gensym/2
 	, dwim_match/2
 	, source_file/1
-	, source_file/2
 	, prolog_load_context/2
 	, current_predicate/2
 	, $defined_predicate/1
 	, predicate_property/2
 	, $predicate_property/2
 	, clause_property/2
-	, clause/2
-	, clause/3
 	, recorda/2
 	, recordz/2
 	, recorded/2
@@ -43,18 +60,22 @@
 	, shell/2
 	, shell/1
 	, shell/0
+	, on_signal/3
+	, current_signal/3
+	, open_shared_object/2
+	, open_shared_object/3
 	, format/1
 	, sformat/2
 	, sformat/3
 	, garbage_collect/0
 	, arithmetic_function/1
         , default_module/2
-	, save_program/1
-	, save_program/2
-	, save/1
 	, absolute_file_name/2
-	, absolute_file_name/3
 	, require/1
+	, call_with_depth_limit/3
+	, length/2
+	, numbervars/3
+	, nb_setval/2				% +Var, +Value
 	]).	
 
 		/********************************
@@ -75,14 +96,19 @@ $map_bits(Pred, ?Name, Old, Old) :-		% ask a bit
 	call(Pred, Name, Bits),
 	Old /\ Bits > 0.
 
-$port_bit(  call, 2'00001).
-$port_bit(  exit, 2'00010).
-$port_bit(  fail, 2'00100).
-$port_bit(  redo, 2'01000).
-$port_bit( unify, 2'10000).
-$port_bit(   all, 2'11111).
-$port_bit(  full, 2'01111).
-$port_bit(  half, 2'01101).
+$port_bit(      call, 2'000000001).
+$port_bit(      exit, 2'000000010).
+$port_bit(      fail, 2'000000100).
+$port_bit(      redo, 2'000001000).
+$port_bit(     unify, 2'000010000).
+$port_bit(     break, 2'000100000).
+$port_bit(  cut_call, 2'001000000).
+$port_bit(  cut_exit, 2'010000000).
+$port_bit( exception, 2'100000000).
+$port_bit(       cut, 2'011000000).
+$port_bit(       all, 2'000111111).
+$port_bit(      full, 2'000101111).
+$port_bit(      half, 2'000101101).
 
 leash(Ports) :-
 	$leash(Old, Old),
@@ -94,24 +120,24 @@ visible(Ports) :-
 	$map_bits($port_bit, Ports, Old, New),
 	$visible(_, New).
 
-$map_style_check(atom,     	   2'00001).
-$map_style_check(singleton, 	   2'00010).
-$map_style_check(dollar,   	   2'00100).
-$map_style_check((discontiguous),  2'01000).
-$map_style_check(string,	   2'10000).
+$map_style_check(atom,     	    2'0000001).
+$map_style_check(singleton, 	    2'0000010).
+$map_style_check(dollar,   	    2'0000100).
+$map_style_check((discontiguous),   2'0001000).
+$map_style_check(dynamic,	    2'0010000).
+$map_style_check(charset,	    2'0100000).
 
+style_check(+string) :- !,
+	set_prolog_flag(double_quotes, string).
+style_check(-string) :- !,
+	set_prolog_flag(double_quotes, codes).
+style_check(?string) :- !,
+	current_prolog_flag(double_quotes, string).
 style_check(Spec) :-
 	$style_check(Old, Old),
 	$map_bits($map_style_check, Spec, Old, New),
 	$style_check(_, New).
 
-please(autoload, Old, New) :- !,
-	flag($enable_autoload, Old, New).
-please(verbose_autoload, Old, New) :- !,
-	flag($verbose_autoload, Old, New).
-please(Key, Old, New) :-
-	$please(Key, Old, New).
-
 :- module_transparent
 	trace/1,
 	trace/2,
@@ -119,6 +145,19 @@ please(Key, Old, New) :-
 	spy/1,
 	nospy/1.
 
+%	prolog:debug_control_hook(+Action)
+%
+%	Allow user-hooks in the Prolog debugger interaction.  See the calls
+%	below for the provided hooks.  We use a single predicate with action
+%	argument to avoid an uncontrolled poliferation of hooks.
+%
+%	TBD: What hooks to provide for trace/[1,2]
+
+:- multifile
+	prolog:debug_control_hook/1.	% +Action
+:- module_transparent
+	prolog:debug_control_hook/1.
+
 trace(Preds) :-
 	trace(Preds, +all).
 
@@ -127,18 +166,27 @@ trace([H|T], Ps) :- !,
 	trace(H, Ps),
 	trace(T, Ps).
 trace(Pred, Ports) :-
-	debug,
+	set_prolog_flag(debug, true),
 	$find_predicate(Pred, Preds),
 	Preds \== [],
-	(   member(Head, Preds),
-		$define_predicate(Head),
-	        $trace(Ports, Head),
-	        show_trace_point(Head),
+	(   $member(Head, Preds),
+	        (   Head = _:_
+		->  QHead0 = Head
+		;   QHead0 = user:Head
+		),
+		$define_predicate(QHead0),
+	        (   predicate_property(QHead0, imported_from(M))
+		->  QHead0 = _:Plain,
+		    QHead = M:Plain
+		;   QHead = QHead0
+		),
+	        $trace(Ports, QHead),
+	        trace_ports(QHead, Tracing),
+	        print_message(informational, trace(QHead, Tracing)),
 	    fail
 	;   true
 	).
 
-
 trace_alias(all,  [trace_call, trace_redo, trace_exit, trace_fail]).
 trace_alias(call, [trace_call]).
 trace_alias(redo, [trace_redo]).
@@ -172,18 +220,17 @@ tag_list([H0|T0], F, [H1|T1]) :-
 	H1 =.. [F, H0],
 	tag_list(T0, F, T1).
 
-
 spy([]) :- !.
 spy([H|T]) :- !,
 	spy(H),
 	spy(T).
+spy(Spec) :-
+	prolog:debug_control_hook(spy(Spec)), !.
 spy(Spec) :-
 	$find_predicate(Spec, Preds),
-	member(Head, Preds),
+	$member(Head, Preds),
 	    $define_predicate(Head),
 	    $spy(Head),
-	    $predicate_name(Head, Name),
-	    $ttyformat('Spy point on ~w~n', [Name]),
 	fail.
 spy(_).
 
@@ -191,64 +238,52 @@ nospy([]) :- !.
 nospy([H|T]) :- !,
 	nospy(H),
 	nospy(T).
+nospy(Spec) :-
+	prolog:debug_control_hook(nospy(Spec)), !.
 nospy(Spec) :-
 	$find_predicate(Spec, Preds),
-	member(Head, Preds),
+	$member(Head, Preds),
 	    $nospy(Head),
-	    $predicate_name(Head, Name),
-	    $ttyformat('Spy point removed from ~w~n', [Name]),
 	fail.
 nospy(_).
 
 nospyall :-
-	current_predicate(_, Module:Head),
-	    $nospy(Module:Head),
+	prolog:debug_control_hook(nospyall),
+	fail.
+nospyall :-
+	spy_point(Head),
+	    $nospy(Head),
 	fail.
 nospyall.
 
 debugging :-
-	$debugging, !,
-	format('Debug mode is on; spy points (see spy/1) on:~n'),
-	$show_spy_points,
-	format('Trace points (see trace/1) on:~n'),
-	show_trace_points.
-
+	prolog:debug_control_hook(debugging), !.
 debugging :-
-	format('Debug mode is off~n').
+	current_prolog_flag(debug, true), !,
+	print_message(informational, debugging(on)),
+	findall(H, spy_point(H), SpyPoints),
+	print_message(informational, spying(SpyPoints)),
+	findall(trace(H,P), trace_point(H,P), TracePoints),
+	print_message(informational, tracing(TracePoints)).
+debugging :-
+	print_message(informational, debugging(off)).
 
-$show_spy_points :-
+spy_point(Module:Head) :-
 	current_predicate(_, Module:Head),
 	$get_predicate_attribute(Module:Head, spy, 1),
-	\+ predicate_property(Module:Head, imported_from(_)),
-	$predicate_name(Module:Head, Name),
-	format('~t~8|~w~n', [Name]),
-	fail.
-$show_spy_points.
+	\+ predicate_property(Module:Head, imported_from(_)).
 
-show_trace_points :-
+trace_point(Module:Head, Ports) :-
 	current_predicate(_, Module:Head),
-	$get_predicate_attribute(Module:Head, trace_any, 1),
-	\+ predicate_property(Module:Head, imported_from(_)),
-	show_trace_point(Module:Head),
-	fail.
-show_trace_points.
+	    $get_predicate_attribute(Module:Head, trace_any, 1),
+	    \+ predicate_property(Module:Head, imported_from(_)),
+	    trace_ports(Module:Head, Ports).
 
-:- module_transparent
-	show_trace_point/1,
-	show_trace_ports/1.
-
-show_trace_point(Head) :-
-	$predicate_name(Head, Name),
-	format('~t~8|~w:', [Name]),
-	show_trace_ports(Head),
-	nl.
-
-show_trace_ports(Head) :-
-	trace_alias(Port, [AttName]),
-	$get_predicate_attribute(Head, AttName, 1),
-	format(' ~w', [Port]),
-	fail.
-show_trace_ports(_).
+trace_ports(Head, Ports) :-
+	findall(Port,
+		(trace_alias(Port, [AttName]),
+		 $get_predicate_attribute(Head, AttName, 1)),
+		Ports).
 
 
 		/********************************
@@ -256,41 +291,24 @@ show_trace_ports(_).
 		*********************************/
 
 concat_atom([A, B], C) :- !,
-	concat(A, B, C).
+	atom_concat(A, B, C).
 concat_atom(L, Atom) :-
 	$concat_atom(L, Atom).
 
 term_to_atom(Term, Atom) :-
-	$term_to_atom(Term, Atom, 0, 0).
-
-atom_to_term(Atom, Term, Bindings) :-
-	var(Bindings),
-	$term_to_atom(Term, Atom, Bindings, 0).
+	atom_to_term(Atom, Term, 0).
 
 int_to_atom(Int, Atom) :-
 	int_to_atom(Int, 10, Atom).
 
-gensym(Base, Atom) :-
-	concat($gs_, Base, Key),
-	flag(Key, Old, Old),
-	succ(Old, New),
-	flag(Key, _, New),
-	concat(Base, New, Atom).
-
 dwim_match(A1, A2) :-
 	dwim_match(A1, A2, _).
 
+
 		/********************************
 		*             SOURCE            *
 		*********************************/
 
-:- module_transparent
-	source_file/2.
-
-source_file(Pred, File) :-
-	current_predicate(_, Pred),
-	$source_file(Pred, File).
-
 source_file(File) :-
 	$time_source_file(File, _).
 source_file(File) :-
@@ -312,13 +330,36 @@ prolog_load_context(module, Module) :-
 prolog_load_context(file, F) :-
 	source_location(F, _).
 prolog_load_context(stream, S) :-
-	current_input(S).
+	source_location(F, _),
+	(   system:$load_input(F, S0)
+	->  S = S0
+	).
 prolog_load_context(directory, D) :-
 	source_location(F, _),
 	file_directory_name(F, D).
 prolog_load_context(term_position, '$stream_position'(0,L,0,0,0)) :-
 	source_location(_, L).
 
+		 /*******************************
+		 *	      CONTROL		*
+		 *******************************/
+
+%	call_with_depth_limit(+Goal, +DepthLimit, -Result)
+%
+%	Try to proof Goal, but fail on any branch exceeding the indicated
+%	depth-limit.  Unify Result with the maximum-reached limit on success,
+%	depth_limit_exceeded if the limit was exceeded and fails otherwise.
+
+:- module_transparent call_with_depth_limit/3.
+
+call_with_depth_limit(G, Limit, Result) :-
+	$depth_limit(Limit, OLimit, OReached),
+	(   catch(G, E, $depth_limit_except(OLimit, OReached, E)),
+	    $depth_limit_true(Limit, OLimit, OReached, Result, Det),
+	    ( Det == ! -> ! ; true )
+	;   $depth_limit_false(OLimit, OReached, Result)
+	).
+
 
 		/********************************
 		*           DATA BASE           *
@@ -353,13 +394,13 @@ current_predicate(Name, Term) :-
 	$c_current_predicate(Name, Term),
 	$defined_predicate(Term), !.
 current_predicate(Name, Term) :-
-	$strip_module(Term, Module, Head),
+	strip_module(Term, Module, Head),
 	default_module(Module, DefModule),
 	$c_current_predicate(Name, DefModule:Head),
 	$defined_predicate(DefModule:Head), !.
 current_predicate(Name, Term) :-
-	flag($enable_autoload, on, on),
-	$strip_module(Term, Module, Head),
+	current_prolog_flag(autoload, true),
+	strip_module(Term, Module, Head),
 	functor(Head, Name, Arity),
 	$find_library(Module, Name, Arity, _LoadModule, _Library), !.
 
@@ -370,8 +411,6 @@ generate_current_predicate(Name, Module, Head) :-
 
 $defined_predicate(Head) :-
 	$get_predicate_attribute(Head, defined, 1), !.
-$defined_predicate(Head) :-
-	$get_predicate_attribute(Head, (dynamic), 1).
 
 :- module_transparent
 	predicate_property/2,
@@ -381,7 +420,7 @@ predicate_property(Pred, Property) :-
 	Property == undefined, !,
 	(   Pred = Module:Head,
 	    var(Module)
-	;   $strip_module(Pred, Module, Head)
+	;   strip_module(Pred, Module, Head)
 	), !,
 	current_module(Module),
 	Term = Module:Head,
@@ -390,39 +429,51 @@ predicate_property(Pred, Property) :-
 	\+ current_predicate(_, Term).
 predicate_property(Pred, Property) :-
 	current_predicate(_, Pred),
-	$predicate_property(Pred, Property).
+	$predicate_property(Property, Pred).
 
-:- index($predicate_property(0, 1)).
-
-$predicate_property(Pred, interpreted) :-
+$predicate_property(interpreted, Pred) :-
 	$get_predicate_attribute(Pred, foreign, 0).
-$predicate_property(Pred, built_in) :-
+$predicate_property(built_in, Pred) :-
 	$get_predicate_attribute(Pred, system, 1).
-$predicate_property(Pred, exported) :-
+$predicate_property(exported, Pred) :-
 	$get_predicate_attribute(Pred, exported, 1).
-$predicate_property(Pred, foreign) :-
+$predicate_property(foreign, Pred) :-
 	$get_predicate_attribute(Pred, foreign, 1).
-$predicate_property(Pred, (dynamic)) :-
+$predicate_property((dynamic), Pred) :-
 	$get_predicate_attribute(Pred, (dynamic), 1).
-$predicate_property(Pred, (volatile)) :-
+$predicate_property((volatile), Pred) :-
 	$get_predicate_attribute(Pred, (volatile), 1).
-$predicate_property(Pred, (multifile)) :-
+$predicate_property((thread_local), Pred) :-
+	$get_predicate_attribute(Pred, (thread_local), 1).
+$predicate_property((multifile), Pred) :-
 	$get_predicate_attribute(Pred, (multifile), 1).
-$predicate_property(Pred, imported_from(Module)) :-
+$predicate_property(imported_from(Module), Pred) :-
 	$get_predicate_attribute(Pred, imported, Module).
-$predicate_property(Pred, transparent) :-
+$predicate_property(transparent, Pred) :-
 	$get_predicate_attribute(Pred, transparent, 1).
-$predicate_property(Pred, indexed(Pattern)) :-
+$predicate_property(indexed(Pattern), Pred) :-
 	$get_predicate_attribute(Pred, indexed, Pattern).
-$predicate_property(Pred, file(File)) :-
+$predicate_property(file(File), Pred) :-
 	source_file(Pred, File).
-$predicate_property(Pred, line_count(LineNumber)) :-
+$predicate_property(line_count(LineNumber), Pred) :-
 	$get_predicate_attribute(Pred, line_count, LineNumber).
-$predicate_property(Pred, notrace) :-
+$predicate_property(notrace, Pred) :-
 	$get_predicate_attribute(Pred, trace, 0).
-$predicate_property(Pred, show_childs) :-
-	$get_predicate_attribute(Pred, system, 1),
-	$get_predicate_attribute(Pred, hide_childs, 0).
+$predicate_property(nodebug, Pred) :-
+	$get_predicate_attribute(Pred, hide_childs, 1).
+$predicate_property(spying, Pred) :-
+	$get_predicate_attribute(Pred, spy, 1).
+$predicate_property(hashed(N), Pred) :-
+	$get_predicate_attribute(Pred, hashed, N),
+	N > 0.
+$predicate_property(references(N), Pred) :-
+	$get_predicate_attribute(Pred, references, N),
+	N \== 0.			% show negative for debugging!
+$predicate_property(number_of_clauses(N), Pred) :-
+	$get_predicate_attribute(Pred, number_of_clauses, N),
+	N \== 0.
+$predicate_property(noprofile, Pred) :-
+	$get_predicate_attribute(Pred, noprofile, 1).
 
 :- index(clause_property(0, 1)).
 
@@ -430,32 +481,12 @@ clause_property(Clause, line_count(LineNumber)) :-
 	$get_clause_attribute(Clause, line_count, LineNumber).
 clause_property(Clause, file(File)) :-
 	$get_clause_attribute(Clause, file, File).
+clause_property(Clause, fact) :-
+	$get_clause_attribute(Clause, fact, true).
+clause_property(Clause, erased) :-
+	$get_clause_attribute(Clause, erased, true).
 
 
-:- module_transparent
-	clause/2,
-	clause/3.
-
-clause(Head, Body, Ref) :-
-	nonvar(Ref), !,
-	$clause(Head, Clause, Ref),
-	$strip_module(Head, _, H),
-	$clause2(H, Body, Clause).
-clause(Head, Body, Ref) :-
-	current_predicate(_, Head),
-	$clause(Head, Clause, Ref),
-	$strip_module(Head, _, H),
-	$clause2(H, Body, Clause).
-
-clause(Head, Body) :-
-	current_predicate(_, Head),
-	$clause(Head, Clause, _),
-	$strip_module(Head, _, H),
-	$clause2(H, Body, Clause).
-
-$clause2(Head, Body, (Head :- Body)) :- !.
-$clause2(Head, true, Head).
-
 recorda(Key, Value) :-
 	recorda(Key, Value, _).
 recordz(Key, Value) :-
@@ -475,9 +506,9 @@ require([N/A|T]) :- !,
 	functor(Head, N, A),
 	$require(Head),
 	require(T).
-require([H|T]) :-
-	$warning('require/1: Illegal predicate specifier: ~w', [H]),
-	require(T).
+require([H|_T]) :-
+	throw(error(type_error(predicate_indicator, H), _)).
+
 
 		/********************************
 		*            MODULES            *
@@ -495,8 +526,8 @@ module(Module) :-
 	current_module(Module), !,
 	$module(_, Module).
 module(Module) :-
-	$break($warning('~w is not a current module', [Module])),
-	$module(_, Module).
+	$module(_, Module),
+	print_message(warning, no_current_module(Module)).
 
 		/********************************
 		*          STATISTICS           *
@@ -523,20 +554,27 @@ statistics :-
 
 	format('~2f seconds cpu time for ~D inferences~n',
 				    [Cputime, Inferences]),
-	format('~D atoms, ~D functors, ~D predicates, ~D modules, ~D byte-codes~n~n',
+	format('~D atoms, ~D functors, ~D predicates, ~D modules, ~D VM-codes~n~n',
 				    [Atoms, Functors, Predicates, Modules, Codes]),
-	format('                      Limit    Allocated       In use~n'),
-	format('Heap         :                  ~t~D~53| Bytes~n',
-	       [Heapused]),
-	format('Local  stack :~t~D~27| ~t~D~40| ~t~D~53| Bytes~n',
+	format('                       Limit    Allocated       In use~n'),
+	(   statistics(heap, Heap),
+	    statistics(heaplimit, HeapLimit)
+	->  format('Heap         :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
+		   [HeapLimit, Heap, Heapused])
+	;   format('Heap         :                  ~t~D~54| Bytes~n',
+		   [Heapused])
+	),
+	format('Local  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
 	       [LocalLimit, Local, LocalUsed]),
-	format('Global stack :~t~D~27| ~t~D~40| ~t~D~53| Bytes~n',
+	format('Global stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
 	       [GlobalLimit, Global, GlobalUsed]),
-	format('Trail  stack :~t~D~27| ~t~D~40| ~t~D~53| Bytes~n',
+	format('Trail  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n~n',
 	       [TrailLimit, Trail, TrailUsed]),
 
 	gc_statistics,
-	shift_statistics.
+	agc_statistics,
+	shift_statistics,
+	thread_statistics.
 
 gc_statistics :-
 	statistics(collections, Collections),
@@ -544,10 +582,19 @@ gc_statistics :-
 	statistics(collected, Collected),
 	statistics(gctime, GcTime),
 
-	format('~n~D garbage collections gained ~D bytes in ~2f seconds.~n',
+	format('~D garbage collections gained ~D bytes in ~2f seconds.~n',
 	       [Collections, Collected, GcTime]).
 gc_statistics.
 
+agc_statistics :-
+	catch(statistics(agc, Agc), _, fail),
+	Agc > 0, !,
+	statistics(agc_gained, Gained),
+	statistics(agc_time, Time),
+	format('~D atom garbage collections gained ~D atoms in ~2f seconds.~n',
+	       [Agc, Gained, Time]).
+agc_statistics.
+
 shift_statistics :-
 	statistics(local_shifts, LS),
 	statistics(global_shifts, GS),
@@ -556,10 +603,20 @@ shift_statistics :-
 	;   GS > 0
 	;   TS > 0
 	), !,
-	format('~nStack shifts: ~D local, ~D global, ~D trail.~n',
+	format('Stack shifts: ~D local, ~D global, ~D trail.~n',
 	       [LS, GS, TS]).
 shift_statistics.
 
+thread_statistics :-
+	current_prolog_flag(threads, true), !,
+	statistics(threads, Active),
+	statistics(threads_created, Created),
+	statistics(thread_cputime, CpuTime),
+	Finished is Created - Active,
+	format('~D threads, ~D finished threads used ~2f seconds.~n',
+	       [Active, Finished, CpuTime]).
+thread_statistics.
+
 
 		/********************************
 		*      SYSTEM INTERACTION       *
@@ -577,6 +634,54 @@ shell :-
 shell :-
 	shell('/bin/sh').
 
+		 /*******************************
+		 *	      SIGNALS		*
+		 *******************************/
+
+:- module_transparent
+	on_signal/3.
+
+on_signal(Signal, Old, New) :-
+	atom(Signal), !,
+	$on_signal(_Num, Signal, Old, New).
+on_signal(Signal, Old, New) :-
+	integer(Signal), !,
+	$on_signal(Signal, _Name, Old, New).
+on_signal(Signal, _Old, _New) :-
+	throw(error(type_error(signal, Signal), on_signal/3)).
+
+current_signal(Name, Id, Handler) :-
+	between(1, 32, Id),
+	$on_signal(Id, Name, Handler, Handler).
+
+
+		 /*******************************
+		 *	      DLOPEN		*
+		 *******************************/
+
+:- module_transparent
+	open_shared_object/2,
+	open_shared_object/3.
+
+dlopen_flag(now,	2'01).		% see pl-load.c for these constants
+dlopen_flag(global,	2'10).		% Solaris only
+
+map_dlflags([], 0).
+map_dlflags([F|T], M) :-
+	map_dlflags(T, M0),
+	dlopen_flag(F, I),
+	M is M0 \/ I.
+
+open_shared_object(File, Flags, Handle) :- % compatibility
+	is_list(Flags), \+ is_list(Handle), !,
+	open_shared_object(File, Handle, Flags).
+open_shared_object(File, Handle, Flags) :-
+	map_dlflags(Flags, Mask),
+	$open_shared_object(File, Handle, Mask).
+
+open_shared_object(File, Handle) :-
+	open_shared_object(File, [], Handle). % use pl-load.c defaults
+
 
 		/********************************
 		*              I/O              *
@@ -594,23 +699,16 @@ sformat(String, Format) :-
 		 *	      FILES		*
 		 *******************************/
 
-%	absolute_file_name(+Term, +Args, -AbsoluteFile)
-
-absolute_file_name(Spec, Args, Path) :-
-	(   select(Args, extensions(Exts), Conditions)
-	->  true
-	;   Conditions = Args,
-	    Exts = ['']
-	),
-	$chk_file(Spec, Exts, Conditions, Path).
-
 %	absolute_file_name(+Term, -AbsoluteFile)
 
 absolute_file_name(Name, Abs) :-
 	atomic(Name), !,
 	$absolute_file_name(Name, Abs).
 absolute_file_name(Term, Abs) :-
-	$chk_file(Term, [''], [], File),
+	$chk_file(Term, [''], [access(read)], File), !,
+	$absolute_file_name(File, Abs).
+absolute_file_name(Term, Abs) :-
+	$chk_file(Term, [''], [], File), !,
 	$absolute_file_name(File, Abs).
 
 
@@ -625,22 +723,6 @@ absolute_file_name(Term, Abs) :-
 garbage_collect :-
 	$garbage_collect(0).
 
-%	save_program(+Name[, +Options])
-%	Save the currently running image in file
-
-save_program(Name) :-
-	save_program(Name, []).
-
-save_program(Name, Options) :-
-	$autoload:clear_library_index,
-	$save_program(Name, Options).
-
-%	save(+File)
-%	Create saved-state in file
-
-save(File) :-
-	save(File, _).
-
 %	arithmetic_function(Spec)
 %	Register a predicate as an arithmetic function.  Takes Name/Arity
 %	and a term as argument.
@@ -649,19 +731,64 @@ save(File) :-
 	arithmetic_function/1.
 
 arithmetic_function(Spec) :-
-	$strip_module(Spec, Module, Term),
+	strip_module(Spec, Module, Term),
 	(   Term = Name/Arity
 	;   functor(Term, Name, Arity)
 	), !,
 	PredArity is Arity + 1,
 	functor(Head, Name, PredArity),
-	$arithmetic_function(Module:Head).
+	$arithmetic_function(Module:Head, 0).
 
 %	default_module(+Me, -Super)
 %	Is true if `Super' is `Me' or a super (auto import) module of `Me'.
 
 default_module(Me, Me).
 default_module(Me, Super) :-
-	$default_module(Me, S, S),
-	S \== [],
+	import_module(Me, S),
 	default_module(S, Super).
+
+
+		 /*******************************
+		 *	 LIST MANIPULATION	*
+		 *******************************/
+
+%	length(?List, ?N)
+%	Is true when N is the length of List.
+
+length(List, Length) :-
+	$length(List, Length), !.		% written in C
+length(List, Length) :-
+	var(Length),
+	length2(List, Length).
+
+length2([], 0).
+length2([_|List], N) :-
+	length2(List, M), 
+	succ(M, N).
+
+
+		 /*******************************
+		 *	       TERM		*
+		 *******************************/
+
+%	numbervars(+Term, +StartIndex, -EndIndex)
+%	
+%	Number all unbound variables in Term   using  $VAR(N), where the
+%	first N is StartIndex and EndIndex is  unified to the index that
+%	will be given to the next variable.
+
+numbervars(Term, From, To) :-
+	numbervars(Term, From, To, []).
+
+
+		 /*******************************
+		 *	       GVAR		*
+		 *******************************/
+
+%	nb_setval(+Name, +Value)
+%	
+%	Bind the non-backtrackable variable Name with a copy of Value
+
+nb_setval(Name, Value) :-
+	duplicate_term(Value, Copy),
+	nb_linkval(Name, Copy).
