@@ -1,30 +1,74 @@
-/*  pl-stream.c,v 1.5 1995/05/23 13:31:31 jan Exp
+/*  $Id: pl-stream.c,v 1.83 2004/09/21 14:22:26 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    See ../LICENCE to find out about your rights.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: SWI-Prolog IO streams
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+#if defined(WIN32) && !defined(__WIN32__)
+#define __WIN32__ 1
+#endif
+
 #ifdef __WIN32__
 #include <uxnt.h>
 #define MD "config/win32.h"
+#include <winsock2.h>
 #endif
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This modules defines the  SWI-Prolog  I/O   streams.  These  streams are
+provided to gain common access to  any   type  of character data: files,
+stdio streams, but also resources, strings, XPCE objects, etc.
+
+MT:
+
+Multithreading is supported through  Slock()   and  Sunlock(). These are
+recursive locks. If a stream handle  might   be  known to another thread
+locking is required. 
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 #ifdef MD
 #include MD
 #else
-#include "config.h"
+#include <config.h>
 #endif
 
-#define _MAKE_DLL
-#undef _export
+#if O_LARGEFILES
+#define _FILE_OFFSET_BITS 64
+#endif
+
+#define PL_KERNEL 1
 #include "pl-stream.h"
 #include <sys/types.h>
+#ifdef HAVE_SYS_TIME_H
+#include <sys/time.h>
+#else
+#include <time.h>
+#endif
 #include <errno.h>
 #ifdef HAVE_MALLOC_H
 #include <malloc.h>
+#else
+#ifdef HAVE_SYS_MALLOC_H
+#include <sys/malloc.h>
+#endif
 #endif
 #include <memory.h>
 #include <string.h>
@@ -32,6 +76,10 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <ctype.h>
+#include <sys/stat.h>
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -50,8 +98,38 @@
 
 #define char_to_int(c)	(0xff & (int)(c))
 
+#define TMPBUFSIZE 256			/* Serror bufsize for Svfprintf() */
+
 int Slinesize = SIO_LINESIZE;		/* Sgets() buffer size */
 
+static int	S__flushbuf(IOSTREAM *s);
+static void	run_close_hooks(IOSTREAM *s);
+
+#define S__fupdatefilepos(s, c) S___fupdatefilepos(s, c)
+
+#ifdef O_PLMT
+#define SLOCK(s)    if ( s->mutex ) recursiveMutexLock(s->mutex)
+#define SUNLOCK(s)  if ( s->mutex ) recursiveMutexUnlock(s->mutex)
+inline int
+STRYLOCK(IOSTREAM *s)
+{ if ( s->mutex &&
+       recursiveMutexTryLock(s->mutex) == MUTEX_BUSY )
+    return FALSE;
+
+  return TRUE;
+}
+#else
+#define SLOCK(s)
+#define SUNLOCK(s)
+#define STRYLOCK(s) (TRUE)
+#endif
+
+#include "pl-error.h"
+extern int 			fatalError(const char *fm, ...);
+extern int 			PL_error(const char *pred, int arity,
+					 const char *msg, int id, ...);
+extern int			PL_handle_signals();
+
 		 /*******************************
 		 *	      BUFFER		*
 		 *******************************/
@@ -81,91 +159,281 @@ S__setbuf(IOSTREAM *s, char *buffer, int size)
   s->limitp   = &s->buffer[s->bufsize];
   s->bufp     = s->buffer;
 
-  return s->bufsize;
+  return size;
 }
 
+
+static int
+S__removebuf(IOSTREAM *s)
+{ if ( s->buffer )
+  { int rval = S__flushbuf(s);
+
+    if ( !(s->flags & SIO_USERBUF) )
+      free(s->buffer);
+    s->bufp = s->limitp = s->buffer = NULL;
+    s->bufsize = 0;
+
+    return rval;
+  }
+
+  return 0;
+}
+
+
+#ifdef DEBUG_IO_LOCKS
+static char *
+Sname(IOSTREAM *s)
+{ if ( s == Serror ) return "error";
+  if ( s == Sinput ) return "input";
+  if ( s == Soutput ) return "output";
+  return "?";
+}
+
+
+#include <execinfo.h>
+#include <string.h>
+
+static void
+print_trace(void)
+{ void *array[3];
+  size_t size;
+  char **strings;
+  size_t i;
+     
+  size = backtrace(array, sizeof(array)/sizeof(void *));
+  strings = backtrace_symbols(array, size);
+     
+  printf(" Stack:");
+  for(i = 1; i < size; i++)
+  { printf(" [%d] %s", i, strings[i]);
+  }
+  printf("\n");
+       
+  free(strings);
+}
+#endif /*DEBUG_IO_LOCKS*/
+
+
+int
+Slock(IOSTREAM *s)
+{ SLOCK(s);
+
+#ifdef DEBUG_IO_LOCKS
+  printf("  Lock: %d: %s: %d locks", PL_thread_self(), Sname(s), s->locks+1);
+  print_trace();
+#endif
+
+  if ( !s->locks++ )
+  { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
+      return S__setbuf(s, NULL, TMPBUFSIZE);
+  }
+
+  return 0;
+}
+
+
+int
+StryLock(IOSTREAM *s)
+{ if ( !STRYLOCK(s) )
+    return -1;
+
+  if ( --s->locks == 0 )
+  { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
+      return S__setbuf(s, NULL, TMPBUFSIZE);
+  }
+
+  return 0;
+}
+
+
+int
+Sunlock(IOSTREAM *s)
+{ int rval = 0;
+
+#ifdef DEBUG_IO_LOCKS
+  printf("Unlock: %d: %s: %d locks", PL_thread_self(), Sname(s), s->locks-1);
+  print_trace();
+#endif
+
+  if ( s->locks )
+  { if ( --s->locks == 0 )
+    { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
+	rval = S__removebuf(s);
+    }
+  }
+
+  SUNLOCK(s);
+
+  return rval;
+}
+
+
 		 /*******************************
 		 *	     FLUSH/FILL		*
 		 *******************************/
 
 static int
 S__flushbuf(IOSTREAM *s)
-{ int size = s->bufp - s->buffer;
+{ int rval = 0;
+  int size;
 
-  if ( (*s->functions->write)(s->handle, s->buffer, size) != size )
-    return -1;
+  while ( (size = s->bufp - s->buffer) > 0 )
+  { int n = (*s->functions->write)(s->handle, s->buffer, size);
 
-  s->bufp = s->buffer;
+    if ( n >= 0 )
+    { rval += n;
+      s->bufp -= n;
+    } else
+    { s->flags |= SIO_FERR;
+      return -1;
+    }
+  }
 
-  return size;
+  return rval;
 }
 
 
 static int
 S__flushbufc(int c, IOSTREAM *s)
-{ if ( s->flags & SIO_NBUF )
-  { char chr = (char)c;
-    
-    if ( (*s->functions->write)(s->handle, &chr, 1) != 1 )
-      return -1;
+{ if ( s->buffer )
+  { if ( S__flushbuf(s) < 0 )
+      c = -1;
+    else
+      *s->bufp++ = (c & 0xff);
   } else
-  { if ( !s->buffer )
-    { if ( S__setbuf(s, NULL, 0) < 0 )
-	return -1;
+  { if ( s->flags & SIO_NBUF )
+    { char chr = (char)c;
+    
+      if ( (*s->functions->write)(s->handle, &chr, 1) != 1 )
+      { s->flags |= SIO_FERR;
+	c = -1;
+      }
     } else
-    { if ( S__flushbuf(s) < 0 )
-	return -1;
+    { if ( S__setbuf(s, NULL, 0) < 0 )
+      { s->flags |= SIO_FERR;
+	c = -1;
+      } else
+	*s->bufp++ = (char)c;
     }
-    *s->bufp++ = (char)c;
   }
 
   return c;
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+S__fillbuf() fills the read-buffer, returning the first character of it.
+It also realises the SWI-Prolog timeout facility.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 S__fillbuf(IOSTREAM *s)
-{ if ( s->flags & (SIO_FEOF|SIO_FERR) )
-    return -1;
+{ int c;
+
+  if ( s->flags & (SIO_FEOF|SIO_FERR) )
+  { s->flags |= SIO_FEOF2;		/* reading past eof */
+    goto error;
+  }
+
+#ifdef HAVE_SELECT
+  s->flags &= ~SIO_TIMEOUT;
+
+  if ( s->timeout >= 0 )
+  { int fd = Sfileno(s);
+
+    if ( fd >= 0 )
+    { fd_set wait;
+      struct timeval time;
+      int rc;
+      
+      time.tv_sec  = s->timeout / 1000;
+      time.tv_usec = (s->timeout % 1000) * 1000;
+      FD_ZERO(&wait);
+#ifdef WIN32
+      FD_SET((SOCKET)fd, &wait);
+#else
+      FD_SET(fd, &wait);
+#endif
+
+      for(;;)
+      { rc = select(fd+1, &wait, NULL, NULL, &time);
+	
+	if ( rc < 0 && errno == EINTR )
+	{ if ( PL_handle_signals() < 0 )
+	  { errno = EPLEXCEPTION;
+	    goto error;
+	  }
+
+	  continue;
+	}
+
+	break;
+      }
+
+      if ( rc == 0 )
+      { s->flags |= (SIO_TIMEOUT|SIO_FERR);
+	goto error;
+      }
+    } else
+    { errno = EPERM;			/* no permission to select */
+      s->flags |= SIO_FERR;
+      goto error;
+    }
+  }
+#endif
+
 
   if ( s->flags & SIO_NBUF )
   { char chr;
     int n;
 
     if ( (n=(*s->functions->read)(s->handle, &chr, 1)) == 1 )
-    { return char_to_int(chr);
+    { c = char_to_int(chr);
+      goto ok;
     } else if ( n == 0 )
-    { s->flags |= SIO_FEOF;
-      return EOF;
+    { if ( !(s->flags & SIO_NOFEOF) )
+	s->flags |= SIO_FEOF;
+      goto error;
     } else
     { s->flags |= SIO_FERR;
-      return -1;			/* error */
+      goto error;			/* error */
     }
   } else
   { int n;
 
     if ( !s->buffer )
     { if ( S__setbuf(s, NULL, 0) < 0 )
-	return -1;
+	goto error;
+      s->limitp = s->buffer;
     }
 
     if ( (n=(*s->functions->read)(s->handle, s->buffer, s->bufsize)) > 0 )
     { s->bufp = s->buffer;
       s->limitp = &s->buffer[n];
-      return char_to_int(*s->bufp++);
+      c = char_to_int(*s->bufp++);
+      goto ok;
     } else
-    { s->bufp = s->buffer;		/* empty the buffer */
-      s->limitp = s->buffer;
-
-      if ( n == 0 )
-      { s->flags |= SIO_FEOF;
-	return EOF;
+    { if ( n == 0 )
+      { if ( !(s->flags & SIO_NOFEOF) )
+	  s->flags |= SIO_FEOF;
+	goto error;
+#ifdef EWOULDBLOCK
+      } else if ( errno == EWOULDBLOCK )
+      { s->bufp = s->buffer;
+	s->limitp = s->buffer;
+	goto error;
+#endif
       } else
       { s->flags |= SIO_FERR;
-	return -1;
+	goto error;
       }
     }
   }
+
+error:
+  c = -1;
+ok:
+  return c;
 }
 
 		 /*******************************
@@ -174,16 +442,32 @@ S__fillbuf(IOSTREAM *s)
 
 
 inline int
-S__fupdatefilepos(IOPOS *p, int c)
-{ if ( p )
+S___fupdatefilepos(IOSTREAM *s, int c)
+{ IOPOS *p;
+
+#if 1
+  if ( s->magic != SIO_MAGIC )
+  { if ( s->magic == SIO_CMAGIC )
+      PL_error(NULL, 0, NULL, ERR_CLOSED_STREAM, s);
+					/* PL_error() should not return */
+    fatalError("Did you load a pre-3.1.2 foreign package?"); 
+  }
+#endif
+
+  if ( (p = s->position) )
   { switch(c)
     { case '\n':
 	p->lineno++;
         p->linepos = 0;
-/*      p->flags &= ~SIO_NOLINEPOS; TBD*/
+	s->flags &= ~SIO_NOLINEPOS;
+#ifdef __WIN32__
+	if ( s->flags & O_TEXT )
+	  p->charno++;			/* writes one extra! */
+#endif
         break;
       case '\r':
 	p->linepos = 0;
+        s->flags &= ~SIO_NOLINEPOS;
 	break;
       case '\b':
 	if ( p->linepos > 0 )
@@ -203,7 +487,6 @@ S__fupdatefilepos(IOPOS *p, int c)
   return c;
 }
 
-
 int
 Sputc(int c, IOSTREAM *s)
 { c &= 0xff;
@@ -212,8 +495,11 @@ Sputc(int c, IOSTREAM *s)
     *s->bufp++ = (char)c;
   else
   { if ( S__flushbufc(c, s) < 0 )
+    { s->lastc = EOF;
       return -1;
+    }
   }
+  s->lastc = c;
 
   if ( c == '\n' && (s->flags & SIO_LBUF) )
   { if ( S__flushbuf(s) < 0 )
@@ -254,7 +540,7 @@ Sungetc(int c, IOSTREAM *s)
 int
 Sputw(int w, IOSTREAM *s)
 { unsigned char *q = (unsigned char *)&w;
-  int n;
+  unsigned int n;
 
   for(n=0; n<sizeof(w); n++)
   { if ( Sputc(*q++, s) < 0 )
@@ -269,7 +555,7 @@ int
 Sgetw(IOSTREAM *s)
 { int w;
   unsigned char *q = (unsigned char *)&w;
-  int n;
+  unsigned int n;
 
   for(n=0; n<sizeof(w); n++)
   { int c;
@@ -297,33 +583,80 @@ Sfread(void *data, int size, int elms, IOSTREAM *s)
     if ( (c = Sgetc(s)) == EOF )
       break;
 
-    *buf++ = c;
+    *buf++ = c & 0xff;
   }
   
-  return chars ? elms : (elms - (chars+size-1)/size);
+  return (size*elms - chars)/size;
 }
 
 
 int
-Sfwrite(void *data, int size, int elms, IOSTREAM *s)
+Sfwrite(const void *data, int size, int elms, IOSTREAM *s)
 { int chars = size * elms;
-  char *buf = data;
+  const char *buf = data;
 
   for( ; chars > 0; chars-- )
-    if ( Sputc(*buf++, s) < 0 )
+  { if ( Sputc(*buf++, s) < 0 )
       break;
+  }
   
-  return chars ? elms : (elms - (chars+size-1)/size);
+  return (size*elms - chars)/size;
 }
 
 
+		 /*******************************
+		 *	       PENDING		*
+		 *******************************/
+
+int
+Sread_pending(IOSTREAM *s, char *buf, int limit, int flags)
+{ int done = 0;
+  int n;
+
+  if ( s->bufp >= s->limitp && (flags & SIO_RP_BLOCK) )
+  { int c = S__fillbuf(s);
+
+    if ( c < 0 )
+      return c;
+
+    buf[0] = c;
+    limit--;
+    done = 1;
+  }
+
+  n = s->limitp - s->bufp;
+  if ( n > limit )
+    n = limit;
+  memcpy(&buf[done], s->bufp, n);
+  s->bufp += n;
+
+  return done+n;
+}
+
+
+
 		 /*******************************
 		 *               FLAGS		*
 		 *******************************/
 
 int
 Sfeof(IOSTREAM *s)
-{ return (s->flags & SIO_FEOF) != 0;
+{ if ( s->flags & SIO_FEOF )
+    return TRUE;
+
+  if ( s->bufp < s->limitp )
+    return FALSE;
+
+  if ( s->flags & SIO_NBUF )
+  { errno = EINVAL;
+    return -1;
+  }
+
+  if ( S__fillbuf(s) == -1 )
+    return TRUE;
+
+  s->bufp--;
+  return FALSE;
 }
     
 
@@ -333,9 +666,15 @@ Sferror(IOSTREAM *s)
 }
     
 
+int
+Sfpasteof(IOSTREAM *s)
+{ return (s->flags & (SIO_FEOF2ERR|SIO_FEOF2)) == (SIO_FEOF2ERR|SIO_FEOF2);
+}
+
+
 void
 Sclearerr(IOSTREAM *s)
-{ s->flags &= ~(SIO_FEOF|SIO_FERR);
+{ s->flags &= ~(SIO_FEOF|SIO_FERR|SIO_FEOF2|SIO_TIMEOUT);
 }
 
 
@@ -346,7 +685,9 @@ Sclearerr(IOSTREAM *s)
 int
 Sflush(IOSTREAM *s)
 { if ( s->buffer && (s->flags & SIO_OUTPUT) )
-    return S__flushbuf(s);
+  { if ( S__flushbuf(s) < 0 )
+      return -1;
+  }
 
   return 0;
 }
@@ -355,14 +696,68 @@ Sflush(IOSTREAM *s)
 		 *	      SEEK		*
 		 *******************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Return the size of the underlying data object.  Should be optimized;
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+long
+Ssize(IOSTREAM *s)
+{ if ( s->functions->control )
+  { long size;
+
+    if ( (*s->functions->control)(s->handle, SIO_GETSIZE, (void *)&size) == 0 )
+      return size;
+  }
+  if ( s->functions->seek )
+  { long here = Stell(s);
+    long end  = Sseek(s, 0, SIO_SEEK_END);
+
+    Sseek(s, here, SIO_SEEK_SET);
+
+    return end;
+  }
+
+  errno = ESPIPE;
+  return -1;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Maybe we should optimise this to become block-aligned?  Or can we leave
 this to read/write?
+
+The first part checks whether repositioning   the  read/write pointer in
+the buffer suffices to achieve the seek.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 long
 Sseek(IOSTREAM *s, long pos, int whence)
-{ if ( !s->functions->seek )
+{ if ( s->limitp > s->buffer )		/* something there */
+  { long now = Stell(s);
+
+    if ( now != -1 )
+    { long rval;
+      char *nbufp = (char *)-1;
+    
+      if ( whence == SIO_SEEK_CUR )
+      { nbufp = s->bufp + pos;
+	rval = now + pos;
+      } else if ( whence == SIO_SEEK_SET )
+      { nbufp = s->bufp + (pos - now);
+	rval = pos;
+      } else
+	rval = -1;			/* should not happen */
+
+      if ( nbufp >= s->buffer && nbufp < s->limitp )
+      { s->bufp = nbufp;
+
+	pos = rval;
+	goto update;
+      }
+    }
+  }
+
+  if ( !s->functions->seek )
   { errno = ESPIPE;
     return -1;
   }
@@ -371,7 +766,6 @@ Sseek(IOSTREAM *s, long pos, int whence)
     
   s->bufp   = s->buffer;
   s->limitp = s->buffer;
-  s->flags &= ~SIO_FEOF;		/* not on eof of file anymore */
 
   if ( whence == SIO_SEEK_CUR )
   { pos += Stell(s);
@@ -379,6 +773,9 @@ Sseek(IOSTREAM *s, long pos, int whence)
   }
   pos = (*s->functions->seek)(s->handle, pos, whence);
 
+update:
+  s->flags &= ~SIO_FEOF;		/* not on eof of file anymore */
+
   if ( s->position )
   { s->flags |= (SIO_NOLINENO|SIO_NOLINEPOS); /* no update this */
     s->position->charno = pos;
@@ -420,8 +817,17 @@ int
 Sclose(IOSTREAM *s)
 { int rval = 0;
 
+  if ( s->magic != SIO_MAGIC )		/* already closed!? */
+  { errno = EINVAL;
+    return -1;
+  }
+
+  SLOCK(s);
+  while(s->locks > 0)			/* remove buffer-locks */
+    rval = Sunlock(s);
+
   if ( s->buffer )
-  { if ( S__flushbuf(s) < 0 )
+  { if ( (s->flags & SIO_OUTPUT) && S__flushbuf(s) < 0 )
       rval = -1;
 
     if ( !(s->flags & SIO_USERBUF) )
@@ -430,10 +836,32 @@ Sclose(IOSTREAM *s)
     s->buffer = NULL;
   }
 
+  s->flags |= SIO_CLOSING;
+#ifdef __WIN32__
+  if ( (s->flags & SIO_ADVLOCK) )
+  { OVERLAPPED ov;
+    HANDLE h = (HANDLE)_get_osfhandle((int)s->handle);
+
+    memset(&ov, 0, sizeof(ov));
+    UnlockFileEx(h, 0, 0, 0xffffffff, &ov);
+    s->flags &= ~SIO_ADVLOCK;
+  }
+#endif
   if ( s->functions->close && (*s->functions->close)(s->handle) < 0 )
     rval = -1;
+  run_close_hooks(s);
 
-  s->magic = 0;
+  SUNLOCK(s);
+
+#ifdef O_PLMT
+  if ( s->mutex )
+  { recursiveMutexDelete(s->mutex);
+    free(s->mutex);
+    s->mutex = NULL;
+  }
+#endif
+
+  s->magic = SIO_CMAGIC;
   if ( !(s->flags & SIO_STATIC) )
     free(s);
 
@@ -454,20 +882,20 @@ Sfgets(char *buf, int n, IOSTREAM *s)
 
     if ( c == EOF )
     { *q = '\0';
-      if ( q > buf )
-	return buf;
-      else
-	return NULL;
+      if ( q == buf )
+	buf = NULL;
+      goto out;
     } else
     { *q++ = c;
       if ( c == '\n' )
       { if ( n > 0 )
 	  *q = '\0';
-	return buf;
+	goto out;
       }
     }
   }
 
+out:
   return buf;
 }
 
@@ -489,7 +917,8 @@ Sgets(char *buf)
 
 int
 Sfputs(const char *q, IOSTREAM *s)
-{ for( ; *q; q++)
+{ 
+  for( ; *q; q++)
   { if ( Sputc(*q, s) < 0 )
       return EOF;
   }
@@ -540,8 +969,8 @@ Svprintf(const char *fm, va_list args)
 }
 
 
-#define OUT(s, c)	do { printed++; \
-			     if ( Sputc((c), (s)) < 0 ) return -1; \
+#define OUTCHR(s, c)	do { printed++; \
+			     if ( Sputc((c), (s)) < 0 ) goto error; \
 			   } while(0)
 #define valdigit(c)	((c) - '0')
 #define A_LEFT	0			/* left-aligned field */
@@ -550,13 +979,23 @@ Svprintf(const char *fm, va_list args)
 int
 Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 { long printed = 0;
+  char buf[TMPBUFSIZE];
+  int tmpbuf;
+
+  SLOCK(s);
+
+  if ( !s->buffer && (s->flags & SIO_NBUF) )
+  { S__setbuf(s, buf, sizeof(buf));
+    tmpbuf = TRUE;
+  } else
+    tmpbuf = FALSE;
 
   while(*fm)
   { if ( *fm == '%' )
     { fm++;
 
       if ( *fm == '%' )
-      { OUT(s, *fm);
+      { OUTCHR(s, *fm);
 	fm++;
 	continue;
       } else
@@ -582,12 +1021,12 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	if ( *fm == '*' )
 	{ has_arg1++;
 	  arg1 = va_arg(args, int);
-	} else if ( isdigit(*fm) )
+	} else if ( isdigit(char_to_int(*fm)) )
 	{ if ( *fm == '0' )
 	    pad = '0';
 	  arg1 = valdigit(*fm);
 	  has_arg1++;
-	  for( fm++; isdigit(*fm); fm++)
+	  for( fm++; isdigit(char_to_int(*fm)); fm++)
 	    arg1 = arg1*10 + valdigit(*fm);
 	}
 	if ( *fm == '.' )
@@ -597,7 +1036,7 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	  { arg2 = va_arg(args, int);
 	  } else
 	  { arg2 = 0;
-	    for( ; isdigit(*fm); fm++)
+	    for( ; isdigit(char_to_int(*fm)); fm++)
 	      arg2 = arg2*10 + valdigit(*fm);
 	  }
 	}
@@ -665,6 +1104,8 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	  }
 	  case 's':
 	    fs = va_arg(args, char *);
+	    if ( !fs )
+	      fs = "(null)";
 	    break;
 	}
 
@@ -675,11 +1116,11 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	  if ( align == A_LEFT )
 	  { int w = 0;
 	    while(*fs)
-	    { OUT(s, *fs++);
+	    { OUTCHR(s, *fs++);
 	      w++;
 	    }
 	    while(w < arg1)
-	    { OUT(s, pad);
+	    { OUTCHR(s, pad);
 	      w++;
 	    }
 	  } else /*if ( align == A_RIGHT ) */
@@ -692,30 +1133,43 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 
 	    w = arg1 - w;
 	    while(w > 0 )
-	    { OUT(s, pad);
+	    { OUTCHR(s, pad);
 	      w--;
 	    }
 	    while(*fs)
-	      OUT(s, *fs++);
+	      OUTCHR(s, *fs++);
 	  }
 	} else
 	{ if ( fs == fbuf )		/* unaligned field, just output */
 	  { while(fs < fe)
-	      OUT(s, *fs++);
+	      OUTCHR(s, *fs++);
 	  } else
 	  { while(*fs)
-	      OUT(s, *fs++);
+	      OUTCHR(s, *fs++);
 	  }
 	}
 	fm++;
       }
+    } else if ( *fm == '\\' && fm[1] )
+    { OUTCHR(s, fm[1]);
+      fm += 2;
     } else
-    { OUT(s, *fm);
+    { OUTCHR(s, *fm);
       fm++;
     }
   }
 
+  if ( tmpbuf )
+  { if ( S__removebuf(s) < 0 )
+      goto error;
+  }
+
+  SUNLOCK(s);
   return printed;
+
+error:
+  SUNLOCK(s);
+  return -1;
 }
 
 
@@ -738,12 +1192,13 @@ Svsprintf(char *buf, const char *fm, va_list args)
   int rval;
 
   s.bufp      = buf;
-  s.limitp    = (char *)0xffffffff;
+  s.limitp    = (char *)(~0L);
   s.buffer    = buf;
   s.flags     = SIO_FBUF|SIO_OUTPUT;
   s.position  = NULL;
   s.handle    = NULL;
   s.functions = NULL;
+  s.mutex     = NULL;
   
   if ( (rval = Svfprintf(&s, fm, args)) >= 0 )
     *s.bufp = '\0';
@@ -755,9 +1210,17 @@ Svsprintf(char *buf, const char *fm, va_list args)
 int
 Svdprintf(const char *fm, va_list args)
 { int rval;
+  IOSTREAM *s = Soutput;
 
-  rval = Svfprintf(Soutput, fm, args);
-  Sflush(Soutput);
+  SLOCK(s);
+  rval = Svfprintf(s, fm, args);
+#if defined(_DEBUG) && defined(WIN32)
+  Sputc('\0', s);
+  s->bufp--;				/* `Unput' */
+  OutputDebugString(s->buffer);
+#endif
+  Sflush(s);
+  SUNLOCK(s);
 
   return rval;
 }
@@ -1053,25 +1516,89 @@ out:
 
 static int
 Sread_file(void *handle, char *buf, int size)
-{ return read((int)handle, buf, size);
+{ long h = (long) handle;
+  int bytes;
+
+  for(;;)
+  { bytes = read((int)h, buf, size);
+
+    if ( bytes == -1 && errno == EINTR )
+    { if ( PL_handle_signals() < 0 )
+      { errno = EPLEXCEPTION;
+	return -1;
+      }
+
+      continue;
+    }
+
+    return bytes;
+  }
 }
 
 
 static int
 Swrite_file(void *handle, char *buf, int size)
-{ return write((int)handle, buf, size);
+{ long h = (long) handle;
+  int bytes;
+
+  for(;;)
+  { bytes = write((int)h, buf, size);
+
+    if ( bytes == -1 && errno == EINTR )
+    { if ( PL_handle_signals() < 0 )
+      { errno = EPLEXCEPTION;
+	return -1;
+      }
+
+      continue;
+    }
+
+    return bytes;
+  }
 }
 
 
 static long
 Sseek_file(void *handle, long pos, int whence)
-{ return lseek((int)handle, pos, whence);
+{ long h = (long) handle;
+
+					/* cannot do EINTR according to man */
+  return lseek((int)h, pos, whence);
 }
 
 
 static int
 Sclose_file(void *handle)
-{ return close((int) handle);
+{ long h = (long) handle;
+  int rc;
+
+  do
+  { rc = close((int) h);
+  }  while ( rc == -1 && errno == EINTR );
+
+  return rc;
+}
+
+
+static int
+Scontrol_file(void *handle, int action, void *arg)
+{ long h = (long) handle;
+  int fd = (int)h;
+
+  switch(action)
+  { case SIO_GETSIZE:
+    { long *rval = arg;
+      struct stat buf;
+
+      if ( fstat(fd, &buf) == 0 )
+      {	*rval = buf.st_size;
+        return 0;
+      }
+      return -1;
+    }
+    default:
+      return -1;
+  }
 }
 
 
@@ -1079,27 +1606,41 @@ IOFUNCTIONS Sfilefunctions =
 { Sread_file,
   Swrite_file,
   Sseek_file,
-  Sclose_file
+  Sclose_file,
+  Scontrol_file
 };
 
 
 IOSTREAM *
 Snew(void *handle, int flags, IOFUNCTIONS *functions)
 { IOSTREAM *s;
+  int fd;
 
   if ( !(s = malloc(sizeof(IOSTREAM))) )
   { errno = ENOMEM;
     return NULL;
   }
   memset((char *)s, 0, sizeof(IOSTREAM));
-  s->magic     = SIO_MAGIC;
-  s->flags     = flags;
-  s->handle    = handle;
-  s->functions = functions;
+  s->magic         = SIO_MAGIC;
+  s->lastc         = EOF;
+  s->flags         = flags;
+  s->handle        = handle;
+  s->functions     = functions;
+  s->timeout       = -1;		/* infinite */
+  s->posbuf.lineno = 1;
   if ( flags & SIO_RECORDPOS )
-  { s->position = &s->posbuf;
-    s->posbuf.lineno = 1;
+    s->position = &s->posbuf;
+#ifdef O_PLMT
+  if ( !(flags & SIO_NOMUTEX) )
+  { if ( !(s->mutex = malloc(sizeof(recursiveMutex))) )
+    { free(s);
+      return NULL;
+    }
+    recursiveMutexInit(s->mutex);
   }
+#endif
+  if ( (fd = Sfileno(s)) >= 0 && isatty(fd) )
+    s->flags |= SIO_ISATTY;
 
   return s;
 }
@@ -1109,79 +1650,211 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
 #define O_BINARY 0
 #endif
 
-IOSTREAM *
-Sopen_file(char *path, char *how)
-{ int fd;
-  int flags;
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Open a file. In addition to the normal  arguments, "lr" means get a read
+(shared-) lock on the file and  "lw"   means  get  an write (exclusive-)
+lock.  How much do we need to test here?
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-  switch(*how)
+
+IOSTREAM *
+Sopen_file(const char *path, const char *how)
+{ int fd;
+  int oflags = 0, flags = SIO_FILE|SIO_TEXT|SIO_RECORDPOS;
+  int op = *how++;
+  long lfd;
+  enum {lnone=0,lread,lwrite} lock = lnone;
+  IOSTREAM *s;
+
+  for( ; *how; how++)
+  { switch(*how)
+    { case 'b':				/* binary */
+	flags &= ~SIO_TEXT;
+        oflags = O_BINARY;
+        break;
+      case 'r':				/* no record */
+	flags &= SIO_RECORDPOS;
+        break;
+      case 'l':				/* lock r: read, w: write */
+	if ( *++how == 'r' )
+	  lock = lread;
+        else if ( *how == 'w' )
+	  lock = lwrite;
+        else
+	{ errno = EINVAL;
+	  return NULL;
+	}
+        break;
+      default:
+	errno = EINVAL;
+        return NULL;
+    }
+  }
+
+#if O_LARGEFILES && defined(O_LARGEFILE)
+  oflags |= O_LARGEFILE
+#endif
+
+  switch(op)
   { case 'w':
-      fd = open(path, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0666);
-      flags = SIO_FILE|SIO_OUTPUT|SIO_RECORDPOS;
+      fd = open(path, O_WRONLY|O_CREAT|O_TRUNC|oflags, 0666);
+      flags |= SIO_OUTPUT;
       break;
     case 'a':
-      fd = open(path, O_WRONLY|O_CREAT|O_APPEND|O_BINARY, 0666);
-      flags = SIO_FILE|SIO_OUTPUT|SIO_RECORDPOS;
+      fd = open(path, O_WRONLY|O_CREAT|O_APPEND|oflags, 0666);
+      flags |= SIO_OUTPUT|SIO_APPEND;
+      break;
+    case 'u':
+      fd = open(path, O_WRONLY|O_CREAT|oflags, 0666);
+      flags |= SIO_OUTPUT|SIO_UPDATE;
       break;
     case 'r':
-      fd = open(path, O_RDONLY|O_BINARY); /* TBD */
-      flags = SIO_FILE|SIO_INPUT|SIO_RECORDPOS;
+      fd = open(path, O_RDONLY|oflags);
+      flags |= SIO_INPUT;
       break;
     default:
       errno = EINVAL;
       return NULL;
   }
 
+
+
   if ( fd < 0 )
     return NULL;
 
-  if ( how[1] == 'b' )
-    flags &= ~SIO_RECORDPOS;
+  if ( lock )
+  { 
+#ifdef FCNTL_LOCKS
+    struct flock buf;
 
-  return Snew((void *)fd, flags, &Sfilefunctions);
+    memset(&buf, 0, sizeof(buf));
+    buf.l_type = (lock == lread ? F_RDLCK : F_WRLCK);
+
+    if ( fcntl(fd, F_SETLKW, &buf) < 0 )
+    { int save = errno;
+      close(fd);
+      errno = save;
+      return NULL;
+    }
+#else					/* we don't have locking */
+#if __WIN32__
+    HANDLE h = (HANDLE)_get_osfhandle(fd);
+    OVERLAPPED ov;
+
+    memset(&ov, 0, sizeof(ov));
+    if ( !LockFileEx(h, (lock == lread ? 0 : LOCKFILE_EXCLUSIVE_LOCK),
+		     0,
+		     0, 0xfffffff,
+		     &ov) )
+    { close(fd);
+      errno = EACCES;			/* TBD: proper error */
+      return NULL;
+    }
+#else
+    close(fd);
+    errno = EINVAL;
+    return NULL;
+#endif
+#endif
+  }
+
+  lfd = (long)fd;
+  s = Snew((void *)lfd, flags, &Sfilefunctions);
+  if ( lock )
+    s->flags |= SIO_ADVLOCK;
+
+  return s;
 }
 
 
 IOSTREAM *
-Sfdopen(int fd, char *type)
+Sfdopen(int fd, const char *type)
 { int flags;
+  long lfd;
 
   if ( fd < 0 )
     return NULL;
 
   if ( *type == 'r' )
-    flags = SIO_FILE|SIO_INPUT;
+    flags = SIO_FILE|SIO_INPUT|SIO_RECORDPOS;
   else
-    flags = SIO_FILE|SIO_OUTPUT;
+    flags = SIO_FILE|SIO_OUTPUT|SIO_RECORDPOS;
 
-  return Snew((void *)fd, flags, &Sfilefunctions);
+  lfd = (long)fd;
+
+  return Snew((void *)lfd, flags, &Sfilefunctions);
 }
 
+/* MT: as long as s is valid, this should be ok
+*/
 
 int
 Sfileno(IOSTREAM *s)
-{ if ( s->flags & SIO_FILE )
-    return (int)s->handle;
-  if ( s->flags & SIO_PIPE )
-    return fileno((FILE *)s->handle);
+{ int n;
 
-  errno = EINVAL;
-  return -1;				/* no file stream */
+  if ( s->flags & SIO_FILE )
+  { long h = (long)s->handle;
+    n = (int)h;
+  } else if ( s->flags & SIO_PIPE )
+  { n = fileno((FILE *)s->handle);
+  } else if ( s->functions->control &&
+	      (*s->functions->control)(s->handle,
+				       SIO_GETFILENO,
+				       (void *)&n)  == 0 )
+  { ;
+  } else
+  { errno = EINVAL;
+    n = -1;				/* no file stream */
+  }
+
+  return n;
 }
 
 
 #define STDIO(n, f) { NULL, NULL, NULL, NULL, \
-		      SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
-		      ((void *)(n)), &Sfilefunctions \
+		      EOF, SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
+		      ((void *)(n)), &Sfilefunctions, \
+		      0, NULL, \
+		      (void (*)(void *))0, NULL, \
+		      -1 \
 		    }
 
+#define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY)
+#define STDIO_STREAMS \
+  STDIO(0, SIO_STDIO|SIO_LBUF|SIO_INPUT|SIO_NOFEOF),	/* Sinput */ \
+  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT), 		/* Soutput */ \
+  STDIO(2, SIO_STDIO|SIO_NBUF|SIO_OUTPUT)		/* Serror */
+
+
 IOSTREAM S__iob[] =
-{ STDIO(0, SIO_FILE|SIO_INPUT|SIO_LBUF|SIO_STATIC),  /* Sinput */
-  STDIO(1, SIO_FILE|SIO_OUTPUT|SIO_LBUF|SIO_STATIC), /* Soutput */
-  STDIO(2, SIO_FILE|SIO_OUTPUT|SIO_NBUF|SIO_STATIC)  /* Serror */
+{ STDIO_STREAMS
 };
 
 
+static const IOSTREAM S__iob0[] =
+{ STDIO_STREAMS
+};
+
+
+void
+SinitStreams()
+{ static int done;
+
+  if ( !done++ )
+  { int i;
+
+    for(i=0; i<=2; i++)
+    { if ( !isatty(i) )
+	S__iob[i].flags &= ~SIO_ISATTY;
+#ifdef O_PLMT
+      S__iob[i].mutex = malloc(sizeof(recursiveMutex));
+      recursiveMutexInit(S__iob[i].mutex);
+#endif
+    }
+  }
+}
+
+
 IOSTREAM *
 S__getiob()
 { return S__iob;
@@ -1192,6 +1865,12 @@ S__getiob()
 		 *******************************/
 
 #ifdef HAVE_POPEN
+#ifdef WIN32
+#include "popen.c"
+
+#define popen(cmd, how) pt_popen(cmd, how)
+#define pclose(fd)	pt_pclose(fd)
+#endif
 
 static int
 Sread_pipe(void *handle, char *buf, int size)
@@ -1221,7 +1900,7 @@ Sclose_pipe(void *handle)
 IOFUNCTIONS Spipefunctions =
 { Sread_pipe,
   Swrite_pipe,
-  NULL,
+  (Sseek_function)0,
   Sclose_pipe
 };
 
@@ -1230,6 +1909,11 @@ IOSTREAM *
 Sopen_pipe(const char *command, const char *type)
 { FILE *fd = popen(command, type);	/* HACK for now */
 
+#if 0
+  Sdprintf("Opening \"%s\", mode \"%s\" --> %p (%d)\n",
+	   command, type, fd, errno);
+#endif
+
   if ( fd )
   { int flags;
 
@@ -1254,6 +1938,8 @@ Sopen_pipe(const char *command, const char *type)
 Memory streams form a replacement for   sprintf(), sscanf() and friends.
 They allow regarding a piece of  (for output) malloc() maintained memory
 to serve as a temporary buffer.
+
+MT: we assume these handles are not passed between threads
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 typedef struct
@@ -1261,16 +1947,25 @@ typedef struct
   long  size;				/* size of buffer */
   int  *sizep;				/* pointer to size */
   long	allocated;			/* allocated size */
-  char *buffer;				/* allocated buffer */
+  char **buffer;			/* allocated buffer */
   int	malloced;			/* malloc() maintained */
 } memfile;
 
 
-static int
-S__memfile_nextsize(int needed)
-{ needed += needed/4;
+void
+Sfree(void *ptr)			/* Windows: must free from same */
+{ free(ptr);				/* DLL */
+}
 
-  return (needed + 255)/256;
+
+static long
+S__memfile_nextsize(long needed)
+{ long size = 512;
+
+  while ( size < needed )
+    size *= 2;
+
+  return size;
 }
 
 
@@ -1278,36 +1973,39 @@ static int
 Swrite_memfile(void *handle, char *buf, int size)
 { memfile *mf = handle;
 
-  if ( mf->here + size >= mf->allocated )
-  { if ( mf->malloced )
-    { long ns = S__memfile_nextsize(mf->here + size);
-      char *nb;
+  if ( mf->here + size + 1 >= mf->allocated )
+  { long ns = S__memfile_nextsize(mf->here + size + 1);
+    char *nb;
 
-      if ( mf->allocated )
-	nb = malloc(ns);
-      else
-	nb = realloc(mf->buffer, ns);
-
-      if ( !nb )
+    if ( mf->allocated == 0 || !mf->malloced )
+    { if ( !(nb = malloc(ns)) )
       { errno = ENOMEM;
 	return -1;
       }
-
-      mf->allocated = ns;
-      mf->buffer = nb;
+      if ( !mf->malloced )
+      { if ( *mf->buffer )
+	  memcpy(nb, *mf->buffer, mf->allocated);
+	mf->malloced = TRUE;
+      }
     } else
-    { errno = ENOMEM;
-      return -1;
+    { if ( !(nb = realloc(*mf->buffer, ns)) )
+      { errno = ENOMEM;
+	return -1;
+      }
     }
+
+    mf->allocated = ns;
+    *mf->buffer = nb;
   }
 
-  memcpy(&mf->buffer[mf->here], buf, size);
+  memcpy(&(*mf->buffer)[mf->here], buf, size);
+  mf->here += size;
 
-  if ( mf->here + size > mf->size )
-  { mf->size = mf->here + size;
+  if ( mf->here > mf->size )
+  { mf->size = mf->here;
     if ( mf->sizep )			/* make externally known */
       *mf->sizep = mf->size;
-    mf->buffer[mf->size] = '\0';
+    (*mf->buffer)[mf->size] = '\0';
   }
 
   return size;
@@ -1324,7 +2022,7 @@ Sread_memfile(void *handle, char *buf, int size)
       size = 0;
   }
   
-  memcpy(buf, &mf->buffer[mf->here], size);
+  memcpy(buf, &(*mf->buffer)[mf->here], size);
   mf->here += size;
 
   return size;
@@ -1381,16 +2079,38 @@ IOFUNCTIONS Smemfunctions =
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Sopenmem(char **buffer, int *size, char* mode)
-    Open an memory area as a stream.  Output streams will automatically
-    resized using realloc() of *size = 0 or the stream is opened with mode
+Sopenmem(char **buffer, int *size, const char* mode)
+    Open a memory area as a stream.  Output streams will automatically
+    resized using realloc() if *size = 0 or the stream is opened with mode
     "wa".
+
+    If the buffer is allocated or enlarged, this is achieved using malloc()
+    or realloc().  In this case the returned buffer should be freed by the
+    caller when done.  Example:
+
+    { char buf[1024];			(don't allocate for small stuff)
+      char *s = buf;
+      IOSTREAM *fd;
+      int size = sizeof(buf);
+
+      fd = Sopenmem(&s, &size, "w");
+      ...
+      Sclose(fd);
+      ...
+      if ( s != buf )			(appearently moved)
+	Sfree(s);
+    }
+
+Note: Its is NOT allows to access   streams  created with this call from
+multiple threads. This is ok for all   usage inside Prolog itself (often
+through tellString()/toldString(). This call is   intented  to use write
+and other output predicates to create strings.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 IOSTREAM *
-Sopenmem(char **buffer, int *sizep, char *mode)
+Sopenmem(char **buffer, int *sizep, const char *mode)
 { memfile *mf = malloc(sizeof(memfile));
-  int flags = SIO_FBUF;
+  int flags = SIO_FBUF|SIO_RECORDPOS|SIO_NOMUTEX;
   int size;
 
   if ( !mf )
@@ -1418,6 +2138,8 @@ Sopenmem(char **buffer, int *sizep, char *mode)
 	mf->malloced = TRUE;
       if ( *buffer )
 	*buffer[0] = '\0';
+      if ( sizep )
+	*sizep = mf->size;
       break;
     default:
       free(mf);
@@ -1427,7 +2149,7 @@ Sopenmem(char **buffer, int *sizep, char *mode)
 
   mf->sizep	= sizep;
   mf->here      = 0;
-  mf->buffer    = *buffer;
+  mf->buffer    = buffer;
 
   return Snew(mf, flags, &Smemfunctions);
 }
@@ -1436,6 +2158,9 @@ Sopenmem(char **buffer, int *sizep, char *mode)
 		 *	      STRINGS		*
 		 *******************************/
 
+/* MT: we assume these handles are not passed between threads
+*/
+
 static int
 Sread_string(void *handle, char *buf, int size)
 { return 0;				/* signal EOF */
@@ -1466,13 +2191,13 @@ Sclose_string(void *handle)
 IOFUNCTIONS Sstringfunctions =
 { Sread_string,
   Swrite_string,
-  NULL,
+  (Sseek_function)0,
   Sclose_string
 };
 
 
 IOSTREAM *
-Sopen_string(IOSTREAM *s, char *buf, int size, char *mode)
+Sopen_string(IOSTREAM *s, char *buf, int size, const char *mode)
 { int flags = SIO_FBUF|SIO_USERBUF;
 
   if ( !s )
@@ -1483,10 +2208,11 @@ Sopen_string(IOSTREAM *s, char *buf, int size, char *mode)
   } else
     flags |= SIO_STATIC;
 
+  memset((char *)s, 0, sizeof(IOSTREAM));
+  s->timeout   = -1;
   s->buffer    = buf;
   s->bufp      = buf;
   s->unbuffer  = buf;
-  s->position  = NULL;
   s->handle    = s;			/* for Sclose_string() */
   s->functions = &Sstringfunctions;
 
@@ -1506,6 +2232,78 @@ Sopen_string(IOSTREAM *s, char *buf, int size, char *mode)
 
   s->flags  = flags;
   s->limitp = &buf[size];
+  s->magic  = SIO_MAGIC;
 
   return s;
 }
+
+#undef S__fupdatefilepos
+
+int
+S__fupdatefilepos(IOSTREAM *s, int c)
+{ return S___fupdatefilepos(s, c);
+}
+
+		 /*******************************
+		 *	       HOOKS		*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This allows external packages (Prolog itself) to monitor the destruction
+of streams.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+typedef struct _close_hook
+{ struct _close_hook *next;
+  void (*hook)(IOSTREAM *s);
+} close_hook;
+
+static close_hook *close_hooks;
+
+static void
+run_close_hooks(IOSTREAM *s)
+{ close_hook *p;
+
+  for(p=close_hooks; p; p = p->next)
+    (*p->hook)(s);
+
+  if ( s->close_hook )
+    (*s->close_hook)(s->closure);
+}
+
+int
+Sclosehook(void (*hook)(IOSTREAM *s))
+{ close_hook *h = malloc(sizeof(*h));
+
+  if ( !h )
+    return -1;
+  h->next = close_hooks;
+  h->hook = hook;
+  close_hooks = h;
+
+  return 0;
+}
+
+
+		 /*******************************
+		 *	       CLEANUP		*
+		 *******************************/
+
+void
+Scleanup(void)
+{ close_hook *p, *next;
+  int i;
+
+  for(p=close_hooks; p; p=next)
+  { next = p->next;
+    free(p);
+  }
+
+  close_hooks = NULL;
+
+  for(i=0; i<=2; i++)
+  { if ( S__iob[i].buffer )
+      free(S__iob[i].buffer);
+    S__iob[i] = S__iob0[i];
+  }
+}
